/**
 * @file Ce fichier regroupe les fonctions fondamentales aux interactions avec le LDAP.
 * C'est ici que tout le filtrage est opéré, au plus bas niveau.
 * Toutes les fonctions écrites ici sont asynchrones et renvoient des Promises ce qui nécessite de les appeler avec la synthaxe
 * un peu particulière `f(args).then(res => ...)` pour exploiter leur résultat.
 * @author hawkspar
 */

import * as basics from './basics';
import fs from 'fs';

//------------------------------------------------------------------------------------------------------------------------
// Fonctions de recherche
//------------------------------------------------------------------------------------------------------------------------

class rechercheModulable {
    /**
     * @class Cette classe contient des fonctions de recherche génériques trop puissantes pour être exportées tel quel.
     * @summary Constructeur vide.
     * @author hawkspar
    */
    constructor() {}

    /**
     * @summary Fonction qui interroge le LDAP et retrouve les groupes (voir LDAP) qui ressemblent
     *  à l'entrée. Etape 0 vers un vrai TOL (Trombino On Line).
     * @desc Cette fonction utilise {@link LDAP.rechercher} mais avec un filtre généré à la volée. 
     * Accepte des champs exacts ou incomplets mais pas approximatifs
     * et ne gère pas l'auto-complete. Cette fonction utilise aussi config.json. MEF Timeout pour
     * des recherches trop vagues. Renvoit une liste d'uid.
     * Elle utilise LDAPEscape pour éviter les injections.
     * @arg {string} input - String entré par l'utilisateur qui ressemble au nom du groupe.
     * @arg {string[]} return_attributes - Liste d'attributs à renvoyer dans le résultat final
     * @return {Promise(string[])} Liste des uid de groupes dont le nom ressemble à l'input
     * @static
     */
    static trouverGroupes(input, return_attributes) {
        return new Promise((resolve, reject) => {
            // Escape de l'input utilisateur par sécurité
            let str=ldapEscape.filter("${txt}", { txt: input});

            // Construction du filtre custom
            let filter= "(|("+config.key_id+"="+str+")"+      // On cherche la valeur exacte
                        "(|("+config.key_id+"=*"+str+")"+     // La valeur finale avec des trucs avant ; wildcard *
                        "(|("+config.key_id+"=*"+str+"*)"+    // La valeur du milieu avec des trucs avant et après
                        "("+config.key_id+"="+str+"*))))";   // La valeur du début avec des trucs après

            // Appel rechercheLDAP avec filtre de l'espace 
            LDAP.rechercher(config.dn_groups, return_attributes, filter).then(res => resolve(res));
        });
    }

    /**
     * @summary Fonction qui renvoit les attributs demandés des paxs validant les critères de recherche. Première étape vers vrai TOL (Trombino On Line).
     * @desc Cette fonction utilise {@link LDAP.rechercher} mais avec un filtre généré à la volée. Accepte des champs exacts ou incomplets pour la plupart des champs
     * mais pas approximatifs et ne gère pas l'auto-complete. MEF Timeout pour des recherches trop vagues. Elle utilise LDAPEscape pour éviter les injections.
     * Utiliser trouverGroupesParTypes pour chaque champ relié à groups.
     * @arg {Object} data - Dictionnaire contenant les données nécessaires à la recherche. Les valeurs sont celles entrées par l'utilisateur et sont par hypothèse
     * comme des sous-parties compactes des valeurs renvoyées. Tous les champs ci-dessous peuvent être indifféremment des listes (par exempl pour chercher un membre
     * de plusieurs groupes) ou des éléments isolés. Si un champ n'est pas pertinent, le mettre à '' ou undefined.
     * @arg {string} data[givenName] - Prénom 
     * @arg {string} data[lastName] - Nom
     * @arg {string} data[nickname] - Surnom
     * @arg {string} data[nationality] - Nationalité (non implémentée pour l'instant, pas de format spécifique)
     * @arg {string} data[promotion] - String de l'année de promo
     * @arg {string} data[phone] - String du numéro de portable
     * @arg {string} data[mail] - Adresse mail
     * @arg {string} data[ips] - Une ou des adresses ip
     * @arg {string} data[school] - Ecole d'appartenance (pour l'instant instable). Doit être exact.
     * @arg {string} data[groups] - Un ou plusieurs groupes (pas de différence entre membre simple et admin). Doit être exact.
     * @arg {string} data[studies] - PA ou autre. Doit être exact.
     * @arg {string} data[sport] - Section sportive ou autre Doit être exact.
     * @arg {string[]} return_attributes - Liste d'attributs à renvoyer dans le résultat final
     * @return {Promise(Object[])} Liste de dictionnaires de profils en cohérence avec l'input avec pour clés les attributs des profils.
     * @static
     */
    static trouverUtilisateurs(data, return_attributes) {
        return new Promise((resolve, reject) => {
            let filter="";
            // Iteration pour chaque champ, alourdissement du filtre selon des trucs prédéfinis dans config encore
            for (var key in data) {
                if ((data[key]!= undefined) & (data[key] != '')) {                    // Si il y a qque chose à chercher pour ce filtre
                    if (!Array.isArray(data[key])) { data[key]=[data[key]]; }        // Gestion d'une liste de valeurs à rechercher
                    // Iteration pour chaque valeur fournie par l'utilisateur
                    data[key].forEach(val => {
                        // Escape de l'input utilisateur
                        let str=ldapEscape.filter("${input}", { input: val});
                        // Traduction en language LDAP
                        let attribute = config.user[key];
                        // Creation du filtre étape par étape
                        filter="(&"+filter+ "(|("+attribute+"="+str+")"+      // On cherche la valeur exacte
                                            "(|("+attribute+"=*"+str+")"+     // La valeur finale avec des trucs avant ; wildcard * (MEF la wildcart ne marche pas pour tous les attributs)
                                            "(|("+attribute+"=*"+str+"*)"+    // La valeur du milieu avec des trucs avant et après
                                            "("+attribute+"="+str+"*)))))";   // La valeur du début avec des trucs après
                    });
                }
            }
            // Appel rechercheLDAP avec filtre de l'espace 
            LDAP.rechercher(config.dn_users, return_attributes, filter).then(res => resolve(res));
        });
    }
}

//------------------------------------------------------------------------------------------------------------------------
// Fonctions intermédiaires TBT
//------------------------------------------------------------------------------------------------------------------------

class test {
    /**
     * @class Cette classe contient des fonctions intermédiaires de calcul utiles mais non exportables. La plupart sont des tests contre le LDAP
     * @summary Constructeur vide.
     * @author hawkspar
     */
    constructor() {}

    /**
     * @summary Cette fonction teste si un utilisateur est membre d'un groupe.
     * @desc Utilise les méthodes statiques listerGroupes et listerMembres de {@link utilisateurAnonyme} mais aussi ajouterMembreGroupe de {@link administrateurConnecte}
     * @param {string} uid - Identifiant de l'utilisateur à tester 
     * @param {string} gid  - Identification du groupe à tester
     * @returns {Promise(boolean)} True si l'utilisateur est membre
     * @static
     */
    static etreMembreGroupe(uid, gid) {
        return new Promise((resolve, reject) => {
            utilisateurAnonyme.listerGroupes(uid).then(lg => {
                utilisateurAnonyme.listerMembres(gid).then(lm => {
                    if (lg.includes(gid) | lm.includes(uid)) {
                        superAdministrateurConnecte.ajouterMembreGroupe(uid, gid);
                        resolve(true);
                    }
                });
            });
        });
    }

    /**
     * @summary Cette fonction teste si un utilisateur est admin d'un groupe.
     * @desc Utilise les méthodes statiques listerAdministrateurs de {@link utilisateurAnonyme} mais aussi ajouterAdministrateurGroupe de {@link administrateurConnecte}
     * @param {string} uid - Identifiant de l'utilisateur à tester 
     * @param {string} gid  - Identification du groupe à tester
     * @returns {Promise(boolean)} True si l'utilisateur est administrateur
     * @static
     */
    static etreAdministrateurGroupe(uid, gid) {
        return new Promise((resolve, reject) => {
            utilisateurAnonyme.listerAdministrateurs(gid).then(la => {
                if (la.includes(uid)) {
                    superAdministrateurConnecte.ajouterAdministrateurGroupe(uid, gid);
                    resolve(true);
                }
            });
        });
    }

    /**
     * @summary Cette fonction teste une valeur d'un attribut (typiquement un identifiant) et le fait évoluer jusqu'à ce qu'il soit unique.
     * @desc Librement adapté de Stack Overflow. Appelle {@link LDAP.rechercher} pour vérifier 
     * qu'il n'y a pas d'autres occurences de cette valeur pour cette attribut
     * dans le dn fourni.
     * @param {string} valeur - Valeur de l'attribut (le plus souvent un identifiant) à tester à cette itération
     * @param {string} attribut - Attribut à tester
     * @param {string} dn - *Domain Name* dans lequel l'attribut doit être unique
     * @param {function} evoluerValeur - Fonction qui prend uniquement en argument l'id courant et 
     * le nombre d'itérations et qui renvoit la prochaine valeur de l'attribut 
     * @param {int} n [0] - Nombre d'itérations (à initialiser à 0)
     * @return {Promise(string)} Valeur unique dans le domaine spécifié de l'attribut spécifié
     * @static
     */
    static assurerUnicite(valeur, attribut, dn, evoluerValeur, n=0) {
        return new Promise((resolve, reject) => { 
            // Recherche d'autres occurences de l'id
            LDAP.rechercher(dn, config.key_id, "("+attribut+"="+valeur+")").then(matches => {
                // On renvoit la valeur si elle est bien unique
                if (matches.length==0) { resolve(valeur); }
                // Sinon, on tente de nouveau notre chance avec la valeur suivante
                else { resolve(test.assurerUnicite(evoluerValeur(valeur, n+1), dn, evoluerValeur, n+1)); }
            });
        });
    }

    /**
     * @summary Cette fonction génère un uid standard, puis le fait évoluer jusqu'à ce qu'il soit unique.
     * @desc Limité à un appel à {@link test.assurerUnicite} avec les bons paramètres, et quelques opérations sur l'uid pour qu'il soit valide (escape, normalisation).
     * @param {string} givenName - Prénom
     * @param {string} lastName - Nom
     * @param {string} promotion - Année de promotion
     * @return {Promise(string)} Valeur unique dans le domaine spécifié de l'attribut spécifié
     * @static
     */
    static genererUid(givenName, lastName, promotion) {
        // Le filtrage évite l'injection de code dans le LDAP, le normalize et lowerCase standardisent le format
        return new Promise((resolve, reject) => {
            test.assurerUnicite(ldapEscape.filter("${uid}",{uid: givenName+'.'+lastName}).toLowerCase().normalize('UFD'), config.key_id, config.dn_users, (id,n) => {
                if (n==1) { id+='.'+ldapEscape.filter("${pr}",{pr: promotion}); }   // Si prénom.nom existe déjà, on rajoute la promo
                else if (n==2) { id+='.'+n-1; }                                     // Puis si prénom.nom.promo existe déjà on passe à nom.prenom.promo .1
                else if (n>2) { id+=+n; }                                           // Ensuite on continue .23, .234, etc...
                return id;
            }).then(id => resolve(id));
        });
    }

    /**
     * @summary Cette fonction génère un gid standard, puis le fait évoluer jusqu'à ce qu'il soit unique.
     * @desc Limité à un appel à {@link test.assurerUnicite} avec les bons paramètres, et quelques opérations sur l'uid pour qu'il soit valide (escape, normalisation).
     * @param {string} name - Nom
     * @return {Promise(string)} Valeur unique dans le domaine spécifié de l'attribut spécifié
     * @static
     */
    static genererGid(name) {
        // Le filtrage évite l'injection de code dans le LDAP, le normalize et lowerCase standardisent le format
        return new Promise((resolve, reject) => {
            test.assurerUnicite(ldapEscape.filter("${id}",{id: name}).toLowerCase().normalize('UFD'), config.key_id, config.dn_groups, (id,n) => {
                if (n==1) { id+='.'+n; }    // Si nom existe déjà, on essaie nom.1
                else if (n>1) { id+=+n; }   // Ensuite on continue .12, .123, etc...
                return id;
            }).then(id => resolve(id));
        });
    }

    /**
     * @summary Cette fonction teste une valeur dummy (0) pour un identifiant numérique puis le fait évoluer aléatoirement (entre 1 et 100 000) jusqu'à ce qu'il soit unique.
     * @param {Object} user - Utilisateur de la forme nécessaire à {@link LDAP.connecter}.
     * @param {string} attribut - Intitulé exact de l'id concerné
     * @param {string} dn - *Domain Name* dans lequel l'attribut doit être unique
     * @return {Promise(string)} Valeur unique dans le domaine spécifié de l'attribut spécifié
     * @static
     */
    static genererIdNum(attribut, dn) {
        return new Promise((resolve,reject) => { test.assurerUnicite("0", attribut, dn, (id,n) => { Math.floor((Math.random() * 100000) + 1).toString(); }).then(id => resolve(id)); });
    }
}